(defconst cobol--imenu-datadiv-sections
  (cobol--with-opt-whitespace-line
   "\\(FILE\\|WORKING-STORAGE\\|LOCAL-STORAGE\\|LINKAGE\\|REPORT\\|SCREEN\\)"
   "[ \t]+SECTION")
  "Regexp which matches a section in the data division.")

(defconst cobol--imenu-indexed-by-re
  ".*INDEXED[ 	]+\\(BY\\)?[ 	]+\\(\\w+\\)"
  "Regexp which matches an INDEXED BY declaration")

(defconst cobol--imenu-procdiv-ignore-re
  (cobol--with-opt-whitespace-line
   (regexp-opt (append cobol-scope-terminators '("EXIT" "GOBACK"))))
  ;; (cobol--with-opt-whitespace-line "\\(end-.*\\|exit\\.\\|goback \\)")
  "Things that match `cobol--procedure-re' which we don't want to add to the
  index.")

(defun cobol--imenu-section-abbreviation (section)
  "Returns an abbreviation of the section name SECTION.

This is to not make the menu entries too wide when there are subprograms whose
  name is prepended to the section."
  (cond ((string= section "FILE")
         " FS")
        ((string= section "WORKING-STORAGE")
         " WSS")
        ((string= section "LOCAL-STORAGE")
         " LSS")
        ((string= section "LINKAGE")
         " LS")
        ((string= section "REPORT")
         " RS")
        ((string= section "SCREEN")
         " SS")))

(defun cobol--imenu-create-datadiv-index ()
  "Create an Imenu sublist item for a section in the data division.

Leaves point at the start of the next section heading in the data
  division, or at the procedure division heading if there were no
  more. Or eob if there is neither."
  (let ((idx-elms nil))
    ;; We stop when we get to the next datadiv section, or when we get to the
    ;; procedure division, or when we hit eob
    (while (and (not (looking-at-p cobol--imenu-datadiv-sections))
                (not (looking-at-p cobol--procedure-division-re))
                (not (eobp)))
      (forward-line 1)
      (unless (looking-at-p comment-re)
        ;; When we're looking at a data item declaration
        (when (looking-at cobol--generic-declaration-re)
          ;; Record the level number and name
          (let ((level (match-string-no-properties 1))
                (name (match-string-no-properties 2)))
            ;; If it is not just a filler, put it in the list
            (when (not (string= (downcase name) "filler"))
              (push (cons (concat level " " name) (match-beginning 2))
                    idx-elms))))
        ;; Pick up INDEXED BY items
        (when (looking-at cobol--imenu-indexed-by-re)
          (push (cons (match-string-no-properties 2)
                      (match-beginning 2))
                idx-elms))))
    ;; Reverse the list so it is in the same order as in the buffer
    (nreverse idx-elms)))


(defun cobol--imenu-create-procdiv-index (program-name)
  "Create imenu index for the procedure division."
  ;; Search for the procedure division start
  (re-search-forward cobol--procedure-division-re nil t)
  ;; Add the procedure division itself to the list
  (let ((idx-elms (list (cons "PD HEADER" (match-beginning 0)))))
    ;; While we're not at the end of buffer or at the end of the next
    ;; subprogram or function
    (while (and (not (eobp))
                (not (looking-at-p cobol--function-end-marker-re)))
      (forward-line 1)
      (when (and (looking-at cobol--procedure-re)
                 (not (looking-at-p cobol--imenu-procdiv-ignore-re)))
        (push (cons (match-string-no-properties 1) (match-beginning 1))
              idx-elms)))
    (nreverse idx-elms)))

(defun cobol--imenu-retrofit-subprogram-name (name index)
  "Retrofit the subprogram name if there were more than one program/function in
the buffer.

INDEX is the list to retrofit in. \"WORKING STORAGE SECTION\" will be replaced
  by \"NAME WSS\" and so on. Returns a new list which is like INDEX except the
  aforementioned transforms have been performed."
  (let ((current (car index))
        (newindex nil))
    (while index
      (let ((current-name (car current)))
        (cond ((string-match cobol--imenu-datadiv-sections current-name)
               (let ((section-name (match-string-no-properties 1 current-name)))
                 (push
                  (cons
                   (concat name
                           (cobol--imenu-section-abbreviation section-name))
                   (cdr current))
                  newindex)))
              ((string-match-p cobol--procedure-division-re current-name)
               (push (cons (concat name " PD") (cdr current)) newindex)))
        (setq index (cdr index))
        (setq current (car index)))
      )
    (nreverse newindex)))
            
(defun cobol--imenu-create-index ()
  "Create an Imenu index for the current buffer.

If there is only one program/function in the buffer each section will have a
menuentry with its full name. If there are more than one programs/functions,
each will have their submenus prefixed by its name and the section names will
abbreviated.

If there are more subprograms or functions in the buffer, there
  will be a submenu listing all the subprograms/functions."
  (goto-char (point-min))
  (let ((case-fold-search t)
        (index nil)                     ;The imenu index
        (subprograms nil)               ;List of subprograms
        (comment-re (if (eq cobol-source-format 'free)
                        cobol--free-form-comment-re
                      cobol--fixed-form-comment-re)))
    ;; While there are more subprograms/functions
    ;; (while (re-search-forward cobol--imenu-identification-division-re nil t)
    ;; (re-search-forward cobol--imenu-program-name nil t)
    (while (re-search-forward cobol--function-id-name-re nil t)
      (let ((subprogram-name (match-string-no-properties 1))
            (subprogram-pos (match-beginning 1)))

        ;; If we get here and `subprograms' has one element, that means there
        ;; are at least two subprograms in the file. Retrofit the last
        ;; subprogram name to the existing menus.
        (when (= (length subprograms) 1)
          (setq index (cobol--imenu-retrofit-subprogram-name
                       (caar subprograms) index)))
        ;; While we're not at the start of the procedure division, search
        ;; forward for a section in the data division
        (while (and (not (looking-at-p cobol--procedure-division-re))
                    (re-search-forward cobol--imenu-datadiv-sections nil t))
          (let ((section (match-string-no-properties 1))
                (subidx (cobol--imenu-create-datadiv-index)))
            (when subidx
              (push
               (cons
                (if subprograms
                    (concat subprogram-name
                            (cobol--imenu-section-abbreviation section))
                  (concat (upcase section) " SECTION"))
                subidx)
               index))))
        ;; We're done indexing the data declarations, now do the procedure
        ;; division
        (let ((procdiv-idx (cobol--imenu-create-procdiv-index subprogram-name)))
          (when procdiv-idx
            (push (cons
                   (if subprograms
                       (concat subprogram-name " PD")
                     "PROCEDURE DIVISION")
                   procdiv-idx)
                  index)))
        (push (cons subprogram-name subprogram-pos) subprograms)))
    ;; If there were multiple programs defined in this file make a submenu with
    ;; their names
    (when (> (length subprograms) 1)
      (push (cons "Subprograms" (nreverse subprograms)) index))
    (nreverse index)))


(defun cobol--find-definition (item)
  "Find the position of the definition of ITEM by searching for it in
`imenu-index-alist'. This is very inefficient!"
  (require 'seq)
  (when (not imenu--index-alist)
    (user-error "Imenu is not active in this buffer"))
  (let* ((res nil)
         (index imenu--index-alist)
         (sublist (cdar index)))
    (while (and (not res) sublist)
      (setq res
            (seq-find
             (lambda (elm)
               (let* ((entrystr (car elm))
                      (identifier
                       (if (string-match cobol--generic-declaration-re entrystr)
                           (match-string-no-properties 2 entrystr)
                         entrystr)))
                 (string= item identifier)))
             sublist))
      (setq index (cdr index))
      (setq sublist (cdar index)))
    (when res (cdr res))))
                          

(defun cobol-goto-definition ()
  "Goto the definition of the identifier under point."
  (interactive)
  (let* ((identifier (symbol-at-point))
         (pos (cobol--find-definition identifier)))
    (when pos
      (push-mark)
      (goto-char pos))))

(defun cobol-peek-definition ()
  "Peek at the definition of the identifier under point."
  (interactive)
  (let* ((identifier (symbol-at-point))
         (pos (cobol--find-definition identifier)))
    (when pos
      (message "%s"
               (save-excursion
                 (goto-char pos)
                 (cobol-back-to-indentation)
                 (buffer-substring (point) (line-end-position)))))))
